/*
 * udpmouse-get.c: An userspace mouse input_dev to receive remote input_event
 *
 * Copyright (c) 2004 Daniele Bellucci <belch@linux.it>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */


#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

#include <linux/input.h>
#include <linux/uinput.h>


static int udp_sock_bind(uint16_t port)
{
    int sock, retval;
    struct sockaddr_in addr;

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
	return -1;
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = INADDR_ANY;

    if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)))
	retval = -1;
    else
	retval = sock;

    return retval;
}


static int uinput_mdev_init(struct uinput_user_dev *dev, int fd)
{
    int retval;

    memset(dev, 0, sizeof(struct uinput_user_dev));
    
    /* set the event types */
    ioctl(fd, UI_SET_EVBIT, EV_REL);
    ioctl(fd, UI_SET_EVBIT, EV_KEY);

    /* set the event codes */
    ioctl(fd, UI_SET_RELBIT, REL_X);
    ioctl(fd, UI_SET_RELBIT, REL_Y);
    ioctl(fd, UI_SET_KEYBIT, BTN_LEFT);
    ioctl(fd, UI_SET_KEYBIT, BTN_RIGHT);
    ioctl(fd, UI_SET_KEYBIT, BTN_MIDDLE);

    retval = write(fd, &dev, sizeof(struct uinput_user_dev));

    if (retval == -1) {
	return -1;
    }
    
    if (ioctl(fd, UI_DEV_CREATE)) {
	return -1;
    }

    return 0;
}

static int uinput_spurious_event(struct input_event *event)
{
    
    if (event->type != EV_REL && event->type != EV_KEY && 
	event->type != EV_SYN) {
	return -1;
    }
    
    if (event->code != REL_X && event->code != REL_Y &&
	event->code != BTN_LEFT && event->code != BTN_RIGHT &&
	event->code != BTN_MIDDLE && event->code != 0) {
	return -1;
    }
    return 0;
}

static int uinput_mdev_event(int fd, struct input_event *event)
{
    return write (fd, event, sizeof(struct input_event));
}


int main (int argc, char **argv)
{
    int fd, sockfd, retval;
    struct uinput_user_dev dev;
    struct input_event event;

    if (argc != 2) {
	fprintf(stderr,"Usage: \"%s\" <port>\n", argv[0]);
	exit(1);
    }

    fd = open("/dev/input/uinput", O_WRONLY);
    if (fd == -1) {
	fprintf(stderr, "%s: open(\"/dev/input/uinput\",...): %s\n", argv[0], 
		strerror(errno));
	exit(1);
    }

    /* create and initialize the uinput_user_dev */
    retval = uinput_mdev_init(&dev, fd);
    if (retval) {
	fprintf(stderr,"%s: uinput_mdev_init(...): %s\n",
		argv[0], strerror(errno));
	exit(1);
    }
    
    /* bind a UDP socket to supplied port number */
    sockfd = udp_sock_bind(atoi(argv[1]));

    while(1)
    {
	/* get an input_event from peer */
	retval = recvfrom(sockfd, &event, sizeof(struct input_event),
			  0, NULL, NULL);
	
	if (retval < 0) {
	    fprintf(stderr,"%s: recvfrom(...): %s\n", argv[0], 
		    strerror(errno));
	    exit(1);
	}

	/* be sure to have received a mouse input_event */
	if (uinput_spurious_event(&event)) {
	    fprintf(stderr,"%s: spurious event received from peer\n", argv[0]);
	    continue;
	}

	/* notify the received input_event */
	retval = uinput_mdev_event(fd, &event);

	if (retval == -1) {
	    fprintf(stderr, "%s: uinput_mdev_event(...): %s\n", argv[0],
		    strerror(errno));
	    exit(1);
	}
    }
}
